#
#  This file is part of Healpy.
#
#  Healpy is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  Healpy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Healpy; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#  For more information about Healpy, see http://code.google.com/p/healpy
#
import numpy as np
import six
import warnings
from . import pixelfunc
from . import sphtfunc
from ._sphtools import rotate_alm

coordname = {"G": "Galactic", "E": "Ecliptic", "C": "Equatorial"}


class ConsistencyWarning(Warning):
    """Warns for a problem in the consistency of data
    """

    pass


if __name__ != "__main__":
    warnings.filterwarnings("always", category=ConsistencyWarning, module=__name__)


class Rotator(object):
    """Rotation operator, including astronomical coordinate systems.

    This class provides tools for spherical rotations. It is meant to be used
    in the healpy library for plotting, and for this reason reflects the
    convention used in the Healpix IDL library.

    Parameters
    ----------
    rot : None or sequence
      Describe the rotation by its euler angle. See :func:`euler_matrix_new`.
    coord : None or sequence of str
      Describe the coordinate system transform. If *rot* is also given, the
      coordinate transform is applied first, and then the rotation.
    inv : bool
      If True, the inverse rotation is defined. (Default: False)
    deg : bool
      If True, angles are assumed to be in degree. (Default: True)
    eulertype : str
      The Euler angle convention used. See :func:`euler_matrix_new`.

    Attributes
    ----------
    mat
    coordin
    coordout
    coordinstr
    coordoutstr
    rots
    coords

    Examples
    --------
    >>> r = Rotator(coord=['G','E'])  # Transforms galactic to ecliptic coordinates
    >>> theta_gal, phi_gal = np.pi/2., 0.
    >>> theta_ecl, phi_ecl = r(theta_gal, phi_gal)  # Apply the conversion
    >>> print(theta_ecl)
    1.66742286715
    >>> print(phi_ecl)
    -1.62596400306
    >>> theta_ecl, phi_ecl = Rotator(coord='ge')(theta_gal, phi_gal) # In one line
    >>> print(theta_ecl)
    1.66742286715
    >>> print(phi_ecl)
    -1.62596400306
    >>> vec_gal = np.array([1, 0, 0]) #Using vectors
    >>> vec_ecl = r(vec_gal)
    >>> print(vec_ecl)
    [-0.05488249 -0.99382103 -0.09647625]
    """

    ErrMessWrongPar = (
        "rot and coord must be single elements or " "sequence of same size."
    )

    def __init__(self, rot=None, coord=None, inv=None, deg=True, eulertype="ZYX"):
        """Create a rotator with given parameters.
        - rot: a float, a tuple of 1,2 or 3 floats or a sequence of tuples.
               If it is a sequence of tuple, it must have the same length as coord.
        - coord: a string or a tuple of 1 or 2 strings or a sequence of tuple
                 If it is a sequence of tuple, it must have same length as rot.
        - inv: whether to use inverse rotation or not
        - deg: if True, angles in rot are assumed in degree (default: True)
        - eulertype: the convention for Euler angles in rot.
        Note: the coord system conversion is applied first, then the rotation.
        """
        rot_is_seq = hasattr(rot, "__len__") and hasattr(rot[0], "__len__")
        coord_is_seq = (
            hasattr(coord, "__len__")
            and hasattr(coord[0], "__len__")
            and type(coord[0]) is not str
        )
        if rot_is_seq and coord_is_seq:
            if len(rot) != len(coord):
                raise ValueError(Rotator.ErrMessWrongPar)
            else:
                rots = rot
                coords = coord
        elif (rot_is_seq or coord_is_seq) and (rot is not None and coord is not None):
            raise ValueError(Rotator.ErrMessWrongPar)
        else:
            rots = [rot]
            coords = [coord]
        inv_is_seq = hasattr(inv, "__len__")
        if inv_is_seq:
            if len(inv) != len(rots):
                raise ValueError("inv must have same length as rot and/or coord")
            invs = inv
        else:
            invs = [inv] * len(rots)
        # check the argument and normalize them
        if eulertype in ["ZYX", "X", "Y"]:
            self._eultype = eulertype
        else:
            self._eultype = "ZYX"
        self._rots = []
        self._coords = []
        self._invs = []
        for r, c, i in zip(rots, coords, invs):
            rn = normalise_rot(r, deg=deg)
            #            if self._eultype in ['X','Y']:
            #                rn[1] = -rn[1]
            cn = normalise_coord(c)
            self._rots.append(rn)  # append(rn) or insert(0, rn) ?
            self._coords.append(cn)  # append(cn) or insert(0, cn) ?
            self._invs.append(bool(i))
        if not self.consistent:
            warnings.warn(
                "The chain of coord system rotations is not consistent",
                category=ConsistencyWarning,
            )
        self._update_matrix()

    def _update_matrix(self):
        self._matrix = np.identity(3)
        self._do_rotation = False
        for r, c, i in zip(self._rots, self._coords, self._invs):
            rotmat, do_rot, rotnorm = get_rotation_matrix(r, eulertype=self._eultype)
            convmat, do_conv, coordnorm = get_coordconv_matrix(c)
            r = np.dot(rotmat, convmat)
            if i:
                r = r.T
            self._matrix = np.dot(self._matrix, r)
            self._do_rotation = self._do_rotation or (do_rot or do_conv)

    def _is_coords_consistent(self):
        for c, i in zip(self._coords, self._invs):
            break
        for cnext, inext in zip(self._coords[1:], self._invs[1:]):
            if c[i] != cnext[not inext]:
                return False
            c, i = cnext, inext
        return True

    consistent = property(
        _is_coords_consistent, doc="consistency of the coords transform chain"
    )

    def __eq__(self, a):
        if type(a) is not type(self):
            return False
        # compare the _rots
        v = [np.allclose(x, y, rtol=0, atol=1e-15) for x, y in zip(self._rots, a._rots)]
        return (
            np.array(v).all()
            and (self._coords == a._coords)
            and (self._invs == a._invs)
        )

    def __call__(self, *args, **kwds):
        """Use the rotator to rotate either spherical coordinates (theta, phi)
        or a vector (x,y,z). You can use lonla keyword to use longitude, latitude
        (in degree) instead of theta, phi (in radian). In this case, returns
        longitude, latitude in degree.

        Accepted forms:

        r(x,y,z)  # x,y,z either scalars or arrays
        r(theta,phi) # theta, phi scalars or arrays
        r(lon,lat,lonlat=True)  # lon, lat scalars or arrays
        r(vec) # vec 1-D array with 3 elements, or 2-D array 3xN
        r(direction) # direction 1-D array with 2 elements, or 2xN array

        Parameters
        ----------
        vec_or_dir : array or multiple arrays
          The direction to rotate. See above for accepted formats.
        lonlat : bool, optional
          If True, assumes the input direction is longitude/latitude in degrees.
          Otherwise, assumes co-latitude/longitude in radians. Default: False
        inv : bool, optional
          If True, applies the inverse rotation. Default: False.
        """
        if kwds.pop("inv", False):
            m = self._matrix.T
        else:
            m = self._matrix
        lonlat = kwds.pop("lonlat", False)
        if len(args) == 1:
            arg = args[0]
            if not hasattr(arg, "__len__") or len(arg) < 2 or len(arg) > 3:
                raise TypeError("Argument must be a sequence of 2 or 3 " "elements")
            if len(arg) == 2:
                return rotateDirection(
                    m, arg[0], arg[1], self._do_rotation, lonlat=lonlat
                )
            else:
                return rotateVector(m, arg[0], arg[1], arg[2], self._do_rotation)
        elif len(args) == 2:
            return rotateDirection(
                m, args[0], args[1], self._do_rotation, lonlat=lonlat
            )
        elif len(args) == 3:
            return rotateVector(m, args[0], args[1], args[2], self._do_rotation)
        else:
            raise TypeError("Either 1, 2 or 3 arguments accepted")

    def __mul__(self, a):
        """Composition of rotation.
        """
        if not isinstance(a, Rotator):
            raise TypeError(
                "A Rotator can only multiply another Rotator "
                "(composition of rotations)"
            )
        rots = self._rots + a._rots
        coords = self._coords + a._coords
        invs = self._invs + a._invs
        return Rotator(rot=rots, coord=coords, inv=invs, deg=False)

    def __rmul__(self, b):
        if not isinstance(b, Rotator):
            raise TypeError(
                "A Rotator can only be multiplied by another Rotator "
                "(composition of rotations)"
            )
        rots = b._rots + self._rots
        coords = b._coords + self._coords
        invs = self._invs + a._invs
        return Rotator(rot=rots, coord=coords, inv=invs, deg=False)

    def __nonzero__(self):
        return self._do_rotation

    def get_inverse(self):
        rots = self._rots[::-1]
        coords = self._coords[::-1]
        invs = [not i for i in self._invs[::-1]]
        return Rotator(rot=rots, coord=coords, inv=invs, deg=False)

    # I = property(get_inverse,doc='Return a new rotator representing the '
    #             'inverse rotation')

    def I(self, *args, **kwds):
        """Rotate the given vector or direction using the inverse matrix.
        rot.I(vec) <==> rot(vec,inv=True)
        """
        kwds["inv"] = True
        return self.__call__(*args, **kwds)

    @property
    def mat(self):
        """The matrix representing the rotation.
        """
        return np.matrix(self._matrix)

    @property
    def coordin(self):
        """The input coordinate system.
        """
        if not self.consistent:
            return None
        for c, i in zip(self._coords, self._invs):
            pass
        return c[i]

    @property
    def coordout(self):
        """The output coordinate system.
        """
        if not self.consistent:
            return None
        for c, i in zip(self._coords, self._invs):
            pass
        return c[not i]

    @property
    def coordinstr(self):
        """The input coordinate system in str.
        """
        return coordname.get(self.coordin, "")

    @property
    def coordoutstr(self):
        """The output coordinate system in str.
        """
        return coordname.get(self.coordout, "")

    @property
    def rots(self):
        """The sequence of rots defining the rotation.
        """
        return self._rots

    @property
    def coords(self):
        """The sequence of coords defining the rotation.
        """
        return self._coords

    def do_rot(self, i):
        """Returns True if rotation is not (close to) identity.
        """
        return not np.allclose(self.rots[i], np.zeros(3), rtol=0., atol=1.e-15)

    def angle_ref(self, *args, **kwds):
        """Compute the angle between transverse reference direction of initial and final frames

        For example, if angle of polarisation is psi in initial frame, it will be psi+angle_ref in final
        frame.

        Parameters
        ----------
        dir_or_vec : array
          Direction or vector (see Rotator.__call__)
        lonlat: bool, optional
          If True, assume input is longitude,latitude in degrees. Otherwise,
          theta,phi in radian. Default: False
        inv : bool, optional
          If True, use the inverse transforms. Default: False

        Returns
        -------
        angle : float, scalar or array
          Angle in radian (a scalar or an array if input is a sequence of direction/vector)
        """
        R = self
        lonlat = kwds.get("lonlat", False)
        inv = kwds.get("inv", False)
        if len(args) == 1:
            arg = args[0]
            if not hasattr(arg, "__len__") or len(arg) < 2 or len(arg) > 3:
                raise TypeError("Argument must be a sequence of 2 or 3 " "elements")
            if len(arg) == 2:
                v = dir2vec(arg[0], arg[1], lonlat=lonlat)
            else:
                v = arg
        elif len(args) == 2:
            v = dir2vec(args[0], args[1], lonlat=lonlat)
        elif len(args) == 3:
            v = args
        else:
            raise TypeError("Either 1, 2 or 3 arguments accepted")
        vp = R(v, inv=inv)
        north_pole = R([0., 0., 1.], inv=inv)
        sinalpha = north_pole[0] * vp[1] - north_pole[1] * vp[0]
        cosalpha = north_pole[2] - vp[2] * np.dot(north_pole, vp)
        return np.arctan2(sinalpha, cosalpha)

    def rotate_alm(self, alm, lmax=None, mmax=None):
        """Rotate Alms with the transform defined in the Rotator object

        see the docstring of the rotate_alm function defined
        in the healpy package, this function **returns** the rotated alms,
        does not rotate in place"""

        rotated_alm = alm.copy()  # rotate_alm works inplace
        rotate_alm(rotated_alm, matrix=self.mat, lmax=lmax, mmax=mmax)
        return rotated_alm

    def rotate_map_alms(self, m, use_pixel_weights=True, lmax=None, mmax=None):
        """Rotate a HEALPix map to a new reference frame in spherical harmonics space

        This is generally the best strategy to rotate/change reference frame of maps.
        If the input map is band-limited, i.e. it can be represented exactly by
        a spherical harmonics transform under a specific lmax, the map rotation
        will be invertible.


        Parameters
        ----------
        m : np.ndarray
            Input map, 1 map is considered I, 2 maps:[Q,U], 3 maps:[I,Q,U]
        use_pixel_weights : bool, optional
            Use pixel weights in map2alm

        Returns
        -------
        m_rotated : np.ndarray
            Map in the new reference frame
        """
        alm = sphtfunc.map2alm(
            m, use_pixel_weights=use_pixel_weights, lmax=lmax, mmax=mmax
        )
        rotated_alm = self.rotate_alm(alm, lmax=lmax, mmax=mmax)
        return sphtfunc.alm2map(
            rotated_alm, lmax=lmax, mmax=mmax, nside=pixelfunc.get_nside(m)
        )

    def rotate_map_pixel(self, m):
        """Rotate a HEALPix map to a new reference frame in pixel space

        It is generally better to rotate in spherical harmonics space, see
        the rotate_map_alms method. A case where pixel space rotation is
        better is for heavily masked maps where the spherical harmonics
        transform is not well defined.
        This function first rotates the pixels centers of the new reference
        frame to the original reference frame, then uses hp.get_interp_val
        to interpolate bilinearly the pixel values, finally fixes Q and U
        polarization by the modification to the psi angle caused by
        the Rotator using Rotator.angle_ref.
        Due to interpolation, this function generally suppresses the signal at
        high angular scales.

        Parameters
        ----------
        m : np.ndarray
            Input map, 1 map is considered I, 2 maps:[Q,U], 3 maps:[I,Q,U]

        Returns
        -------
        m_rotated : np.ndarray
            Map in the new reference frame

        """
        if pixelfunc.maptype(m) == 0:  # a single map is converted to a list
            m = [m]
        npix = len(m[0])
        nside = pixelfunc.npix2nside(npix)
        theta_pix_center, phi_pix_center = pixelfunc.pix2ang(
            nside=nside, ipix=np.arange(npix)
        )

        # Rotate the pixels center of the new reference frame to the original frame
        theta_pix_center_rot, phi_pix_center_rot = self.I(
            theta_pix_center, phi_pix_center
        )

        # Interpolate the original map to the pixels centers in the new ref frame
        m_rotated = [
            pixelfunc.get_interp_val(each, theta_pix_center_rot, phi_pix_center_rot)
            for each in m
        ]

        # Rotate polarization
        if len(m_rotated) > 1:
            # Create a complex map from QU  and apply the rotation in psi due to the rotation
            # Slice from the end of the array so that it works both for QU and IQU
            L_map = (m_rotated[-2] + m_rotated[-1] * 1j) * np.exp(
                1j * 2 * self.angle_ref(theta_pix_center_rot, phi_pix_center_rot)
            )

            # Overwrite the Q and U maps with the correct values
            m_rotated[-2] = np.real(L_map)
            m_rotated[-1] = np.imag(L_map)
        else:
            m_rotated = m_rotated[0]

        return m_rotated

    def __repr__(self):
        return (
            "[ "
            + ", ".join([str(self._coords), str(self._rots), str(self._invs)])
            + " ]"
        )

    __str__ = __repr__


################################################################
#
#     Helpers function for rotation
#     used in the Rotator class.


def rotateVector(rotmat, vec, vy=None, vz=None, do_rot=True):
    """Rotate a vector (or a list of vectors) using the rotation matrix
    given as first argument.

    Parameters
    ----------
    rotmat : float, array-like shape (3,3)
      The rotation matrix
    vec : float, scalar or array-like
      The vector to transform (shape (3,) or (3,N)),
      or x component (scalar or shape (N,)) if vy and vz are given
    vy : float, scalar or array-like, optional
      The y component of the vector (scalar or shape (N,))
    vz : float, scalar or array-like, optional
      The z component of the vector (scalar or shape (N,))
    do_rot : bool, optional
      if True, really perform the operation, if False do nothing.

    Returns
    -------
    vec : float, array
      The component of the rotated vector(s).

    See Also
    --------
    Rotator
    """
    if vy is None and vz is None:
        if do_rot:
            return np.tensordot(rotmat, vec, axes=(1, 0))
        else:
            return vec
    elif vy is not None and vz is not None:
        if do_rot:
            return np.tensordot(rotmat, np.array([vec, vy, vz]), axes=(1, 0))
        else:
            return vec, vy, vz
    else:
        raise TypeError("You must give either vec only or vec, vy " "and vz parameters")


def rotateDirection(rotmat, theta, phi=None, do_rot=True, lonlat=False):
    """Rotate the vector described by angles theta,phi using the rotation matrix
    given as first argument.

    Parameters
    ----------
    rotmat : float, array-like shape (3,3)
      The rotation matrix
    theta : float, scalar or array-like
      The angle theta (scalar or shape (N,))
      or both angles (scalar or shape (2, N)) if phi is not given.
    phi : float, scalar or array-like, optionnal
      The angle phi (scalar or shape (N,)).
    do_rot : bool, optional
      if True, really perform the operation, if False do nothing.
    lonlat : bool
      If True, input angles are assumed to be longitude and latitude in degree,
      otherwise, they are co-latitude and longitude in radians.

    Returns
    -------
    angles : float, array
      The angles of describing the rotated vector(s).

    See Also
    --------
    Rotator
    """
    vx, vy, vz = rotateVector(rotmat, dir2vec(theta, phi, lonlat=lonlat), do_rot=do_rot)
    return vec2dir(vx, vy, vz, lonlat=lonlat)


def vec2dir(vec, vy=None, vz=None, lonlat=False):
    """Transform a vector to angle given by theta,phi.

    Parameters
    ----------
    vec : float, scalar or array-like
      The vector to transform (shape (3,) or (3,N)),
      or x component (scalar or shape (N,)) if vy and vz are given
    vy : float, scalar or array-like, optional
      The y component of the vector (scalar or shape (N,))
    vz : float, scalar or array-like, optional
      The z component of the vector (scalar or shape (N,))
    lonlat : bool, optional
      If True, return angles will be longitude and latitude in degree,
      otherwise, angles will be longitude and co-latitude in radians (default)

    Returns
    -------
    angles : float, array
      The angles (unit depending on *lonlat*) in an array of
      shape (2,) (if scalar input) or (2, N)

    See Also
    --------
    :func:`dir2vec`, :func:`pixelfunc.ang2vec`, :func:`pixelfunc.vec2ang`
    """
    if np.any(np.isnan(vec)):
        return np.nan, np.nan
    if vy is None and vz is None:
        vx, vy, vz = vec
    elif vy is not None and vz is not None:
        vx = vec
    else:
        raise TypeError("You must either give both vy and vz or none of them")
    r = np.sqrt(vx ** 2 + vy ** 2 + vz ** 2)
    ang = np.empty((2, r.size))
    ang[0, :] = np.arccos(vz / r)
    ang[1, :] = np.arctan2(vy, vx)
    if lonlat:
        ang = np.degrees(ang)
        np.negative(ang[0, :], ang[0, :])
        ang[0, :] += 90.
        return ang[::-1, :].squeeze()
    else:
        return ang.squeeze()


def dir2vec(theta, phi=None, lonlat=False):
    """Transform a direction theta,phi to a unit vector.

    Parameters
    ----------
    theta : float, scalar or array-like
      The angle theta (scalar or shape (N,))
      or both angles (scalar or shape (2, N)) if phi is not given.
    phi : float, scalar or array-like, optionnal
      The angle phi (scalar or shape (N,)).
    lonlat : bool
      If True, input angles are assumed to be longitude and latitude in degree,
      otherwise, they are co-latitude and longitude in radians.

    Returns
    -------
    vec : array
      The vector(s) corresponding to given angles, shape is (3,) or (3, N).

    See Also
    --------
    :func:`vec2dir`, :func:`pixelfunc.ang2vec`, :func:`pixelfunc.vec2ang`
    """
    if phi is None:
        theta, phi = theta
    if lonlat:
        lon, lat = theta, phi
        theta, phi = np.pi / 2. - np.radians(lat), np.radians(lon)
    ct, st, cp, sp = np.cos(theta), np.sin(theta), np.cos(phi), np.sin(phi)
    vec = np.empty((3, ct.size), np.float64)
    vec[0, :] = st * cp
    vec[1, :] = st * sp
    vec[2, :] = ct
    return vec.squeeze()


def angdist(dir1, dir2, lonlat=False):
    """Returns the angular distance between dir1 and dir2.

    Parameters
    ----------
    dir1, dir2 : float, array-like
      The directions between which computing the angular distance.
      Angular if len(dir) == 2 or vector if len(dir) == 3.
      See *lonlat* for unit
    lonlat : bool, scalar or sequence
      If True, angles are assumed to be longitude and latitude in degree,
      otherwise they are interpreted as colatitude and longitude in radian.
      If a sequence, lonlat[0] applies to dir1 and lonlat[1] applies to dir2.

    Returns
    -------
    angles : float, scalar or array-like
      The angle(s) between dir1 and dir2 in radian.

    Examples
    --------
    >>> import healpy as hp
    >>> hp.rotator.angdist([.2,0], [.2, 1e-6])
    array([  1.98669331e-07])
    """
    if hasattr(lonlat, "__len__") and len(lonlat) == 2:
        lonlat1, lonlat2 = lonlat
    else:
        lonlat1 = lonlat2 = lonlat
    dir1 = np.asarray(dir1)
    dir2 = np.asarray(dir2)
    if dir1.ndim == 2:
        if dir1.shape[0] == 2:  # theta, phi -> vec
            vec1 = dir2vec(dir1, lonlat=lonlat1)
        else:
            vec1 = np.reshape(dir1, (3, -1))
            vec1 = normalize_vec(vec1)
    elif dir1.ndim == 1:
        if dir1.shape[0] == 2:  # theta, phi -> vec
            vec1 = np.reshape(dir2vec(dir1, lonlat=lonlat1), (3, 1))
        else:
            vec1 = np.reshape(dir1, (3, 1))
            vec1 = normalize_vec(vec1)
    if dir2.ndim == 2:
        if dir2.shape[0] == 2:  # theta, phi -> vec
            vec2 = dir2vec(dir2, lonlat=lonlat2)
        else:
            vec2 = np.reshape(dir2, (3, -1))
            vec2 = normalize_vec(vec2)
    elif dir2.ndim == 1:
        if dir2.shape[0] == 2:  # theta, phi -> vec
            vec2 = np.reshape(dir2vec(dir2, lonlat=lonlat2), (3, 1))
        else:
            vec2 = np.reshape(dir2, (3, 1))
            vec2 = normalize_vec(vec2)
    # compute vec product
    vec_prod = np.sqrt((np.cross(vec1.T, vec2.T) ** 2).sum(axis=1))
    # compute scalar product
    scal_prod = (vec1 * vec2).sum(axis=0)

    return np.arctan2(vec_prod, scal_prod)


def normalize_vec(vec):
    """Normalize the vector(s) *vec* (in-place if it is a ndarray).

    Parameters
    ----------
    vec : float, array-like of shape (D,) or (D, N)
      The D-vector(s) to normalize.

    Returns
    -------
    vec_normed : float, array
      Normalized vec, shape (D,) or (D, N)
    """
    vec = np.array(vec, np.float64)
    r = np.sqrt(np.sum(vec ** 2, axis=0))
    vec /= r
    return vec


#######################################################
#
#   Manage the coord system conventions
#


def check_coord(c):
    """Check if parameter is a valid coord system.
    Raise a TypeError exception if it is not, otherwise returns the normalized
    coordinate system name.
    """
    if c is None:
        return c
    if not isinstance(c, six.string_types):
        raise TypeError(
            "Coordinate must be a string (G[alactic],"
            " E[cliptic], C[elestial]"
            " or Equatorial=Celestial)"
        )
    if c[0].upper() == "G":
        x = "G"
    elif c[0].upper() == "E" and c != "Equatorial":
        x = "E"
    elif c[0].upper() == "C" or c == "Equatorial":
        x = "C"
    else:
        raise ValueError(
            "Wrong coordinate (either G[alactic],"
            " E[cliptic], C[elestial]"
            " or Equatorial=Celestial)"
        )
    return x


def normalise_coord(coord):
    """Normalise the coord argument.
    Coord sys are either 'E','G', 'C' or 'X' if undefined.

    Input: either a string or a sequence of string.

    Output: a tuple of two strings, each being one of the norm coord sys name
            above.

    eg, 'E' -> ['E','E'], ['Ecliptic','G'] -> ['E','G']
    None -> ['X','X'] etc.
    """
    coord_norm = []
    if coord is None:
        coord = (None, None)
    coord = tuple(coord)
    if len(coord) > 2:
        raise TypeError(
            "Coordinate must be a string (G[alactic],"
            " E[cliptic] or C[elestial])"
            " or a sequence of 2 strings"
        )
    for x in coord:
        coord_norm.append(check_coord(x))
    if len(coord_norm) < 2:
        coord_norm.append(coord_norm[0])
    return tuple(coord_norm)


def normalise_rot(rot, deg=False):
    """Return rot possibly completed with zeroes to reach size 3.
   If rot is None, return a vector of 0.
   If deg is True, convert from degree to radian, otherwise assume input
   is in radian.
   """
    if deg:
        convert = np.pi / 180.
    else:
        convert = 1.
    if rot is None:
        rot = np.zeros(3)
    else:
        rot = np.array(rot, np.float64).flatten() * convert
        rot.resize(3)
    return rot


def get_rotation_matrix(rot, deg=False, eulertype="ZYX"):
    """Return the rotation matrix corresponding to angles given in rot.

   Usage: matrot,do_rot,normrot = get_rotation_matrix(rot)

   Input:
      - rot: either None, an angle or a tuple of 1,2 or 3 angles
             corresponding to Euler angles.
   Output:
      - matrot: 3x3 rotation matrix
      - do_rot: True if rotation is not identity, False otherwise
      - normrot: the normalized version of the input rot.
   """
    rot = normalise_rot(rot, deg=deg)
    if not np.allclose(rot, np.zeros(3), rtol=0., atol=1.e-15):
        do_rot = True
    else:
        do_rot = False
    if eulertype == "X":
        matrot = euler_matrix_new(rot[0], -rot[1], rot[2], X=True)
    elif eulertype == "Y":
        matrot = euler_matrix_new(rot[0], -rot[1], rot[2], Y=True)
    else:
        matrot = euler_matrix_new(rot[0], -rot[1], rot[2], ZYX=True)

    return matrot, do_rot, rot


def get_coordconv_matrix(coord):
    """Return the rotation matrix corresponding to coord systems given
   in coord.

   Usage: matconv,do_conv,normcoord = get_coordconv_matrix(coord)

   Input:
      - coord: a tuple with initial and final coord systems.
               See normalise_coord.
   Output:
      - matconv: the euler matrix for coord sys conversion
      - do_conv: True if matconv is not identity, False otherwise
      - normcoord: the tuple of initial and final coord sys.

   History: adapted from CGIS IDL library.
   """

    coord_norm = normalise_coord(coord)

    if coord_norm[0] == coord_norm[1]:
        matconv = np.identity(3)
        do_conv = False
    else:
        eps = 23.452294 - 0.0130125 - 1.63889E-6 + 5.02778E-7
        eps = eps * np.pi / 180.

        # ecliptic to galactic
        e2g = np.array(
            [
                [-0.054882486, -0.993821033, -0.096476249],
                [0.494116468, -0.110993846, 0.862281440],
                [-0.867661702, -0.000346354, 0.497154957],
            ]
        )

        # ecliptic to equatorial
        e2q = np.array(
            [
                [1., 0., 0.],
                [0., np.cos(eps), -1. * np.sin(eps)],
                [0., np.sin(eps), np.cos(eps)],
            ]
        )

        # galactic to ecliptic
        g2e = np.linalg.inv(e2g)

        # galactic to equatorial
        g2q = np.dot(e2q, g2e)

        # equatorial to ecliptic
        q2e = np.linalg.inv(e2q)

        # equatorial to galactic
        q2g = np.dot(e2g, q2e)

        if coord_norm == ("E", "G"):
            matconv = e2g
        elif coord_norm == ("G", "E"):
            matconv = g2e
        elif coord_norm == ("E", "C"):
            matconv = e2q
        elif coord_norm == ("C", "E"):
            matconv = q2e
        elif coord_norm == ("C", "G"):
            matconv = q2g
        elif coord_norm == ("G", "C"):
            matconv = g2q
        else:
            raise ValueError("Wrong coord transform :", coord_norm)
        do_conv = True

    return matconv, do_conv, coord_norm


###################################################
##                                               ##
##                euler functions                ##
##                                               ##
######                                      #######


def euler(ai, bi, select, FK4=0):
    """
   NAME:
       euler
   PURPOSE:
       Transform between Galactic, celestial, and ecliptic coordinates.
   EXPLANATION:
       Use the procedure ASTRO to use this routine interactively

   CALLING SEQUENCE:
        EULER, AI, BI, AO, BO, [ SELECT, /FK4, SELECT = ]

   INPUTS:
         AI - Input Longitude in DEGREES, scalar or vector.  If only two
                 parameters are supplied, then  AI and BI will be modified
                 to contain the output longitude and latitude.
         BI - Input Latitude in DEGREES

   OPTIONAL INPUT:
         SELECT - Integer (1-6) specifying type of coordinate
                  transformation.

        SELECT   From          To        |   SELECT      From         To
         1     RA-Dec (2000)  Galactic   |     4       Ecliptic     RA-Dec
         2     Galactic       RA-DEC     |     5       Ecliptic    Galactic
         3     RA-Dec         Ecliptic   |     6       Galactic    Ecliptic

        If not supplied as a parameter or keyword, then EULER will prompt
        for the value of SELECT
        Celestial coordinates (RA, Dec) should be given in equinox J2000
        unless the /FK4 keyword is set.
   OUTPUTS:
         AO - Output Longitude in DEGREES
         BO - Output Latitude in DEGREES

   INPUT KEYWORD:
         /FK4 - If this keyword is set and non-zero, then input and output
               celestial and ecliptic coordinates should be given in
               equinox B1950.
         /SELECT  - The coordinate conversion integer (1-6) may
                    alternatively be specified as a keyword
   NOTES:
         EULER was changed in December 1998 to use J2000 coordinates as the
         default, ** and may be incompatible with earlier versions***.
   REVISION HISTORY:
         Written W. Landsman,  February 1987
         Adapted from Fortran by Daryl Yentis NRL
         Converted to IDL V5.0   W. Landsman   September 1997
         Made J2000 the default, added /FK4 keyword
          W. Landsman December 1998
         Add option to specify SELECT as a keyword W. Landsman March 2003
         Converted to python by K. Ganga December 2007
   """

    # npar = N_params()
    #  if npar LT 2 then begin
    #     print,'Syntax - EULER, AI, BI, A0, B0, [ SELECT, /FK4, SELECT= ]'
    #     print,'    AI,BI - Input longitude,latitude in degrees'
    #     print,'    AO,BO - Output longitude, latitude in degrees'
    #     print,'    SELECT - Scalar (1-6) specifying transformation type'
    #     return
    #  endif

    PI = np.pi
    twopi = 2.0 * PI
    fourpi = 4.0 * PI
    deg_to_rad = 180.0 / PI
    #
    # ;   J2000 coordinate conversions are based on the following constants
    # ;   (see the Hipparcos explanatory supplement).
    # ;  eps = 23.4392911111 # Obliquity of the ecliptic
    # ;  alphaG = 192.85948d           Right Ascension of Galactic North Pole
    # ;  deltaG = 27.12825d            Declination of Galactic North Pole
    # ;  lomega = 32.93192d            Galactic longitude of celestial equator
    # ;  alphaE = 180.02322d           Ecliptic longitude of Galactic North Pole
    # ;  deltaE = 29.811438523d        Ecliptic latitude of Galactic North Pole
    # ;  Eomega  = 6.3839743d          Galactic longitude of ecliptic equator
    #
    if FK4 == 1:

        equinox = "(B1950)"
        psi = [
            0.57595865315,
            4.9261918136,
            0.00000000000,
            0.0000000000,
            0.11129056012,
            4.7005372834,
        ]
        stheta = [
            0.88781538514,
            -0.88781538514,
            0.39788119938,
            -0.39788119938,
            0.86766174755,
            -0.86766174755,
        ]
        ctheta = [
            0.46019978478,
            0.46019978478,
            0.91743694670,
            0.91743694670,
            0.49715499774,
            0.49715499774,
        ]
        phi = [
            4.9261918136,
            0.57595865315,
            0.0000000000,
            0.00000000000,
            4.7005372834,
            0.11129056012,
        ]
    else:

        equinox = "(J2000)"
        psi = [
            0.57477043300,
            4.9368292465,
            0.00000000000,
            0.0000000000,
            0.11142137093,
            4.71279419371,
        ]
        stheta = [
            0.88998808748,
            -0.88998808748,
            0.39777715593,
            -0.39777715593,
            0.86766622025,
            -0.86766622025,
        ]
        ctheta = [
            0.45598377618,
            0.45598377618,
            0.91748206207,
            0.91748206207,
            0.49714719172,
            0.49714719172,
        ]
        phi = [
            4.9368292465,
            0.57477043300,
            0.0000000000,
            0.00000000000,
            4.71279419371,
            0.11142137093,
        ]
    #
    i = select - 1  # IDL offset
    a = ai / deg_to_rad - phi[i]
    b = bi / deg_to_rad
    sb = np.sin(b)
    cb = np.cos(b)
    cbsa = cb * np.sin(a)
    b = -stheta[i] * cbsa + ctheta[i] * sb
    # bo    = math.asin(where(b<1.0, b, 1.0)*deg_to_rad)
    bo = np.arcsin(b) * deg_to_rad
    #
    a = np.arctan2(ctheta[i] * cbsa + stheta[i] * sb, cb * np.cos(a))
    ao = np.fmod((a + psi[i] + fourpi), twopi) * deg_to_rad
    return ao, bo


def euler_matrix_new(a1, a2, a3, X=True, Y=False, ZYX=False, deg=False):
    """
   NAME:
     euler_matrix_new

   PURPOSE:
     computes the Euler matrix of an arbitrary rotation described
     by 3 Euler angles
     correct bugs present in Euler_Matrix

   CALLING SEQUENCE:
     result = euler_matrix_new (a1, a2, a3 [,X, Y, ZYX, DEG ])

   INPUTS:
     a1, a2, a3 = Euler angles, scalar
                  (in radian by default, in degree if DEG is set)
                  all the angles are measured counterclockwise
                  correspond to x, y, zyx-conventions (see Goldstein)
                  the default is x

   KEYWORD PARAMETERS:
      DEG : if set the angle are measured in degree

      X :       rotation a1 around original Z
                rotation a2 around interm   X
                rotation a3 around final    Z
                DEFAULT,  classical mechanics convention

      Y :       rotation a1 around original Z
                rotation a2 around interm   Y
                rotation a3 around final    Z
                quantum mechanics convention (override X)

      ZYX :     rotation a1 around original Z
                rotation a2 around interm   Y
                rotation a3 around final    X
                aeronautics convention (override X)

      * these last three keywords are obviously mutually exclusive *

   OUTPUTS:
      result is a 3x3 matrix

   USAGE:
     if vec is an Nx3 array containing N 3D vectors,
     vec # euler_matrix_new(a1,a2,a3,/Y) will be the rotated vectors


   MODIFICATION HISTORY:
      March 2002, EH, Caltech, rewritting of euler_matrix

      convention   euler_matrix_new           euler_matrix
     X:       M_new(a,b,c,/X)  =  M_old(-a,-b,-c,/X) = Transpose( M_old(c, b, a,/X))
     Y:       M_new(a,b,c,/Y)  =  M_old(-a, b,-c,/Y) = Transpose( M_old(c,-b, a,/Y))
   ZYX:       M_new(a,b,c,/Z)  =  M_old(-a, b,-c,/Z)
   """

    t_k = 0
    if ZYX:
        t_k = t_k + 1
    # if X:   t_k = t_k + 1
    if Y:
        t_k = t_k + 1
    if t_k > 1:
        raise ValueError("Choose either X, Y or ZYX convention")

    convert = 1.0
    if deg:
        convert = np.pi / 180.

    c1 = np.cos(a1 * convert)
    s1 = np.sin(a1 * convert)
    c2 = np.cos(a2 * convert)
    s2 = np.sin(a2 * convert)
    c3 = np.cos(a3 * convert)
    s3 = np.sin(a3 * convert)

    if ZYX:
        m1 = np.array([[c1, -s1, 0], [s1, c1, 0], [0, 0, 1]])  # around   z

        m2 = np.array([[c2, 0, s2], [0, 1, 0], [-s2, 0, c2]])  # around   y

        m3 = np.array([[1, 0, 0], [0, c3, -s3], [0, s3, c3]])  # around   x

    elif Y:
        m1 = np.array([[c1, -s1, 0], [s1, c1, 0], [0, 0, 1]])  # around   z

        m2 = np.array([[c2, 0, s2], [0, 1, 0], [-s2, 0, c2]])  # around   y

        m3 = np.array([[c3, -s3, 0], [s3, c3, 0], [0, 0, 1]])  # around   z

    else:
        m1 = np.array([[c1, -s1, 0], [s1, c1, 0], [0, 0, 1]])  # around   z

        m2 = np.array([[1, 0, 0], [0, c2, -s2], [0, s2, c2]])  # around   x

        m3 = np.array([[c3, -s3, 0], [s3, c3, 0], [0, 0, 1]])  # around   z

    M = np.dot(m3.T, np.dot(m2.T, m1.T))

    return M


def coordsys2euler_zyz(coord):
    """Return the ZYZ Euler angles corresponding to a rotation between
    the two coordinate systems.  The angles can be used in rotate_alm.

    Parameters
    ----------
    coord : a tuple with initial and final coord systems.
      See normalise_coord.

    Returns
    -------
    psi, theta, phi :  floats
       The Euler angles defining a ZYZ rotation between the coordinate
        systems.

    Examples
    --------
    >>> np.array(coordsys2euler_zyz('GE'))
    array([ 1.45937485,  1.05047962, -3.14119347])
    >>> np.array(coordsys2euler_zyz('CG'))
    array([-0.22443941,  1.09730866,  2.14556934])
    >>> np.array(coordsys2euler_zyz('E'))
    array([ 0.,  0.,  0.])
    """

    coord_norm = normalise_coord(coord)

    if coord_norm[0] == coord_norm[1]:
        psi, theta, phi = np.zeros(3)
    else:
        # Convert basis vectors

        xin, yin, zin = np.eye(3)
        rot = Rotator(coord=coord_norm)
        xout, yout, zout = rot([xin, yin, zin]).T

        # Normalize

        xout /= np.sqrt(np.dot(xout, xout))
        yout /= np.sqrt(np.dot(yout, yout))
        zout /= np.sqrt(np.dot(zout, zout))

        # Get the angles

        psi = np.arctan2(yout[2], -xout[2])
        theta = np.arccos(np.dot(zin, zout))
        phi = np.arctan2(zout[1], zout[0])

    return psi, theta, phi



#
#  This file is part of Healpy.
#
#  Healpy is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  Healpy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Healpy; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#  For more information about Healpy, see http://code.google.com/p/healpy
#
"""This module provides classes for some spherical projection.
To be used when calling SphereProjAxes class.

SphericalProj : a virtual class (do nothing). Just a template for derived 
                (useful) classes

GnomonicProj : Gnomonic projection

AzimuthalProj : Azimuthal equidistant or Lambert azimuthal equal-area projection
"""

from . import rotator as R
import numpy as np
from . import pixelfunc
from .pixelfunc import UNSEEN

pi = np.pi
dtor = np.pi / 180.


class SphericalProj(object):
    """
    This class defines functions for spherical projection.
    
    This class contains class method for spherical projection computation. It 
    should not be instantiated. It should be inherited from and methods should
    be overloaded for desired projection.
    """

    name = "None"

    def __init__(self, rot=None, coord=None, flipconv=None, **kwds):
        self.rotator = R.Rotator(rot=rot, coord=None, eulertype="ZYX")
        self.coordsys = R.Rotator(coord=coord).coordout
        self.coordsysstr = R.Rotator(coord=coord).coordoutstr
        self.set_flip(flipconv)
        self.set_proj_plane_info(**kwds)

    def set_proj_plane_info(self, **kwds):
        allNone = True
        for v in kwds.values():
            if v is not None:
                allNone = False
        if not allNone:
            self._arrayinfo = dict(kwds)
        else:
            self._arrayinfo = None

    def get_proj_plane_info(self):
        return self._arrayinfo

    arrayinfo = property(
        get_proj_plane_info, doc="Dictionary with information on the projection array"
    )

    def __eq__(self, a):
        if type(a) is not type(self):
            return False
        return (self.rotator == a.rotator) and (self.coordsys == a.coordsys)

    def ang2xy(self, theta, phi=None, lonlat=False, direct=False):
        """From angular direction to position in the projection plane (%s).

        Input:
          - theta: if phi is None, theta[0] contains theta, theta[1] contains phi
          - phi  : if phi is not None, theta,phi are direction
          - lonlat: if True, angle are assumed in degree, and longitude, latitude
          - flipconv is either 'astro' or 'geo'. None will be default.
        Return:
          - x, y: position in %s plane.
        """
        pass

    def vec2xy(self, vx, vy=None, vz=None, direct=False):
        """From unit vector direction to position in the projection plane (%s).

        Input:
          - vx: if vy and vz are None, vx[0],vx[1],vx[2] defines the unit vector.
          - vy,vz: if defined, vx,vy,vz define the unit vector
          - lonlat: if True, angle are assumed in degree, and longitude, latitude
          - flipconv is either 'astro' or 'geo'. None will be default.

        Return:
          - x, y: position in %s plane.
        """
        pass

    def xy2ang(self, x, y=None, lonlat=False, direct=False):
        """From position in the projection plane to angular direction (%s).

        Input:
          - x : if y is None, x[0], x[1] define the position in %s plane.
          - y : if defined, x,y define the position in projection plane.
          - lonlat: if True, angle are assumed in degree, and longitude, latitude
          - flipconv is either 'astro' or 'geo'. None will be default.

        Return:
          - theta, phi : angular direction.
        """
        pass

    def xy2vec(self, x, y=None, direct=False):
        """From position in the projection plane to unit vector direction (%s).

        Input:
          - x : if y is None, x[0], x[1] define the position in %s plane.
          - y : if defined, x,y define the position in projection plane.
          - lonlat: if True, angle are assumed in degree, and longitude, latitude
          - flipconv is either 'astro' or 'geo'. None will be default.

        Return:
          - theta, phi : angular direction.
        """
        pass

    def xy2ij(self, x, y=None):
        """From position in the projection plane to image array index (%s).

        Input:
          - x : if y is None, x[0], x[1] define the position in %s plane.
          - y : if defined, x,y define the position in projection plane.
          - projinfo : additional projection information.

        Return:
          - i,j : image array indices.
        """
        pass

    def ij2xy(self, i=None, j=None):
        """From image array indices to position in projection plane (%s).

        Input:
          - if i and j are None, generate arrays of i and j as input
          - i : if j is None, i[0], j[1] define array indices in %s image.
          - j : if defined, i,j define array indices in image.
          - projinfo : additional projection information.

        Return:
          - x,y : position in projection plane.
        """
        pass

    def projmap(self, map, vec2pix_func, rot=None, coord=None):
        """Create an array containing the projection of the map.

        Input:
          - vec2pix_func: a function taking theta,phi and returning pixel number
          - map: an array containing the spherical map to project,
                 the pixelisation is described by vec2pix_func
        Return:
          - a 2D array with the projection of the map.

        Note: the Projector must contain information on the array.
        """
        x, y = self.ij2xy()
        if np.__version__ >= "1.1":
            matype = np.ma.core.MaskedArray
        else:
            matype = np.ma.array
        if type(x) is matype and x.mask is not np.ma.nomask:
            w = x.mask == False
        else:
            w = slice(None)
        img = np.zeros(x.shape, np.float64) - np.inf
        vec = self.xy2vec(np.asarray(x[w]), np.asarray(y[w]))
        vec = (R.Rotator(rot=rot, coord=self.mkcoord(coord))).I(vec)
        pix = vec2pix_func(vec[0], vec[1], vec[2])
        # support masked array for map, or a dictionnary (for explicit pixelisation)
        if isinstance(map, matype) and map.mask is not np.ma.nomask:
            mpix = map[pix]
            mpix[map.mask[pix]] = UNSEEN
        elif isinstance(map, dict):
            is_pix_seen = np.in1d(pix, map.keys()).reshape(pix.shape)
            is_pix_unseen = ~is_pix_seen
            mpix = np.zeros_like(img[w])
            mpix[is_pix_unseen] = UNSEEN
            pix_seen = pix[is_pix_seen]
            iterable = (map[p] for p in pix_seen)
            mpix[is_pix_seen] = np.fromiter(iterable, mpix.dtype, count=pix_seen.size)
        else:
            mpix = map[pix]
        img[w] = mpix
        return img

    def set_flip(self, flipconv):
        """flipconv is either 'astro' or 'geo'. None will be default.
        
        With 'astro', east is toward left and west toward right. 
        It is the opposite for 'geo'
        """
        if flipconv is None:
            flipconv = "astro"  # default
        if flipconv == "astro":
            self._flip = -1
        elif flipconv == "geo":
            self._flip = 1
        else:
            raise ValueError("flipconv must be 'astro', 'geo' or None for default.")

    def get_extent(self):
        """Get the extension of the projection plane.

        Return:
          extent = (left,right,bottom,top)
        """
        pass

    def get_fov(self):
        """Get the field of view in degree of the plane of projection

        Return:
          fov: the diameter in radian of the field of view
        """
        return 2. * pi

    def get_center(self, lonlat=False):
        """Get the center of the projection.

        Input:
          - lonlat : if True, will return longitude and latitude in degree,
                     otherwise, theta and phi in radian
        Return:
          - theta,phi or lonlat depending on lonlat keyword
        """
        lon, lat = np.asarray(self.rotator.rots[0][0:2]) * 180 / pi
        if lonlat:
            return lon, lat
        else:
            return pi / 2. - lat * dtor, lon * dtor

    def mkcoord(self, coord):
        if self.coordsys is None:
            return (coord, coord)
        elif coord is None:
            return (self.coordsys, self.coordsys)
        elif type(coord) is str:
            return (coord, self.coordsys)
        else:
            return (tuple(coord)[0], self.coordsys)


class GnomonicProj(SphericalProj):
    """This class provides class methods for Gnomonic projection.
    """

    name = "Gnomonic"

    def __init__(self, rot=None, coord=None, xsize=None, ysize=None, reso=None, **kwds):
        super(GnomonicProj, self).__init__(
            rot=rot, coord=coord, xsize=xsize, ysize=ysize, reso=reso, **kwds
        )

    def set_proj_plane_info(self, xsize=200, ysize=None, reso=1.5):
        if xsize is None:
            xsize = 200
        if ysize is None:
            ysize = xsize
        if reso is None:
            reso = 1.5
        super(GnomonicProj, self).set_proj_plane_info(
            xsize=xsize, ysize=ysize, reso=reso
        )

    def vec2xy(self, vx, vy=None, vz=None, direct=False):
        if not direct:
            vec = self.rotator(vx, vy, vz)
        elif vy is None and vz is None:
            vec = vx
        elif vy is not None and vz is not None:
            vec = vx, vy, vz
        else:
            raise ValueError("vy and vz must be both defined or both not defined")
        flip = self._flip
        mask = np.asarray(vec[0]) <= 0.
        w = np.where(mask == False)
        if not mask.any():
            mask = np.ma.nomask
        if not hasattr(vec[0], "__len__"):
            if mask is not np.ma.nomask:
                x = np.nan
                y = np.nan
            else:
                x = flip * vec[1] / vec[0]
                y = vec[2] / vec[0]
        else:
            x = np.zeros(vec[0].shape) + np.nan
            y = np.zeros(vec[0].shape) + np.nan
            x[w] = flip * vec[1][w] / vec[0][w]
            y[w] = vec[2][w] / vec[0][w]
        return x, y

    vec2xy.__doc__ = SphericalProj.ang2xy.__doc__ % (name, name)

    def xy2vec(self, x, y=None, direct=False):
        flip = self._flip
        if y is None:
            x, y = x
        x, y = np.asarray(x), np.asarray(y)
        rm1 = 1. / np.sqrt(1. + x ** 2 + y ** 2)
        vec = (rm1, flip * rm1 * x, rm1 * y)
        if not direct:
            return self.rotator.I(vec)
        else:
            return vec

    xy2vec.__doc__ = SphericalProj.xy2vec.__doc__ % (name, name)

    def ang2xy(self, theta, phi=None, lonlat=False, direct=False):
        vec = R.dir2vec(theta, phi, lonlat=lonlat)
        return self.vec2xy(vec, direct=direct)

    ang2xy.__doc__ = SphericalProj.ang2xy.__doc__ % (name, name)

    def xy2ang(self, x, y=None, lonlat=False, direct=False):
        return R.vec2dir(self.xy2vec(x, y, direct=direct), lonlat=lonlat)

    xy2ang.__doc__ = SphericalProj.xy2ang.__doc__ % (name, name)

    def xy2ij(self, x, y=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for " "this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        ysize = int(self.arrayinfo["ysize"])
        reso = self.arrayinfo["reso"]
        if y is None:
            x, y = x
        dx = reso / 60. * dtor
        xc, yc = 0.5 * (xsize - 1), 0.5 * (ysize - 1)
        j = np.around(xc + x / dx).astype(np.long)
        i = np.around(yc + y / dx).astype(np.long)
        return i, j

    xy2ij.__doc__ = SphericalProj.xy2ij.__doc__ % (name, name)

    def ij2xy(self, i=None, j=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for " "this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        ysize = int(self.arrayinfo["ysize"])
        reso = self.arrayinfo["reso"]
        dx = reso / 60. * dtor
        xc, yc = 0.5 * (xsize - 1), 0.5 * (ysize - 1)
        if i is None and j is None:
            idx = np.outer(np.ones(ysize), np.arange(xsize))
            x = (idx - xc) * dx  # astro= '-' sign, geo '+' sign
            idx = np.outer(np.arange(ysize), np.ones(xsize))
            y = (idx - yc) * dx  # (idx-yc) * dx
        elif i is not None and j is not None:
            x = (np.asarray(j) - xc) * dx
            y = (np.asarray(i) - yc) * dx  # (asarray(i)-yc) * dx
        elif i is not None and j is None:
            i, j = i
            x = (np.asarray(j) - xc) * dx
            y = (np.asarray(i) - yc) * dx  # (i-yc) * dx
        else:
            raise TypeError("Wrong parameters")
        return x, y

    ij2xy.__doc__ = SphericalProj.ij2xy.__doc__ % (name, name)

    def get_extent(self):
        xsize, ysize = self.arrayinfo["xsize"], self.arrayinfo["ysize"]
        left, bottom = self.ij2xy(0, 0)
        right, top = self.ij2xy(ysize - 1, xsize - 1)
        return (left, right, bottom, top)

    def get_fov(self):
        vx, vy, vz = self.xy2vec(self.ij2xy(0, 0), direct=True)
        a = np.arccos(vx)
        return 2. * a


class MollweideProj(SphericalProj):
    """This class provides class methods for Mollweide projection.
    """

    name = "Mollweide"
    __molldata = []

    def __init__(self, rot=None, coord=None, xsize=800, **kwds):
        self.__initialise_data()
        super(MollweideProj, self).__init__(rot=rot, coord=coord, xsize=xsize, **kwds)

    def set_proj_plane_info(self, xsize):
        super(MollweideProj, self).set_proj_plane_info(xsize=xsize)

    def vec2xy(self, vx, vy=None, vz=None, direct=False):
        if not direct:
            theta, phi = R.vec2dir(self.rotator(vx, vy, vz))
        else:
            theta, phi = R.vec2dir(vx, vy, vz)
        flip = self._flip
        X, Y = MollweideProj.__molldata
        # set phi in [-pi,pi]
        phi = (phi + pi) % (2 * pi) - pi
        lat = pi / 2. - theta
        A = MollweideProj.__lininterp(X, Y, lat)
        x = flip * 2. / pi * phi * np.cos(A)
        y = np.sin(A)
        return x, y

    vec2xy.__doc__ = SphericalProj.vec2xy.__doc__ % (name, name)

    def xy2vec(self, x, y=None, direct=False):
        flip = self._flip
        if y is None:
            x, y = x
        mask = np.asarray(x) ** 2 / 4. + np.asarray(y) ** 2 > 1.
        w = np.where(mask == False)
        if not mask.any():
            mask = np.ma.nomask
        if not hasattr(x, "__len__"):
            if mask is not np.ma.nomask:
                return np.nan, np.nan, np.nan
            else:
                s = np.sqrt((1 - y) * (1 + y))
                a = np.arcsin(y)
                z = 2. / pi * (a + y * s)
                phi = flip * pi / 2. * x / np.maximum(s, 1.e-6)
                sz = np.sqrt((1 - z) * (1 + z))
                vec = sz * np.cos(phi), sz * np.sin(phi), z
                if not direct:
                    return self.rotator.I(vec)
                else:
                    return vec
        else:
            vec = (
                np.zeros(x.shape) + np.nan,
                np.zeros(x.shape) + np.nan,
                np.zeros(x.shape) + np.nan,
            )
            s = np.sqrt((1 - y[w]) * (1 + y[w]))
            a = np.arcsin(y[w])
            vec[2][w] = 2. / pi * (a + y[w] * s)
            phi = flip * pi / 2. * x[w] / np.maximum(s, 1.e-6)
            sz = np.sqrt((1 - vec[2][w]) * (1 + vec[2][w]))
            vec[0][w] = sz * np.cos(phi)
            vec[1][w] = sz * np.sin(phi)
            if not direct:
                return self.rotator.I(vec)
            else:
                return vec

    xy2vec.__doc__ = SphericalProj.xy2vec.__doc__ % (name, name)

    def ang2xy(self, theta, phi=None, lonlat=False, direct=False):
        return self.vec2xy(R.dir2vec(theta, phi, lonlat=lonlat), direct=direct)

    ang2xy.__doc__ = SphericalProj.ang2xy.__doc__ % (name, name)

    def xy2ang(self, x, y=None, lonlat=False, direct=False):
        vec = self.xy2vec(x, y, direct=direct)
        return R.vec2dir(vec, lonlat=lonlat)

    xy2ang.__doc__ = SphericalProj.xy2ang.__doc__ % (name, name)

    def xy2ij(self, x, y=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for " "this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        ysize = xsize // 2
        if y is None:
            x, y = x
        xc, yc = (xsize - 1.) / 2., (ysize - 1.) / 2.
        if hasattr(x, "__len__"):
            j = np.around(x * xc / 2. + xc).astype(np.long)
            i = np.around(yc + y * yc).astype(np.long)
            mask = x ** 2 / 4. + y ** 2 > 1.
            if not mask.any():
                mask = np.ma.nomask
            j = np.ma.array(j, mask=mask)
            i = np.ma.array(i, mask=mask)
        else:
            if x ** 2 / 4. + y ** 2 > 1.:
                i, j = np.nan, np.nan
            else:
                j = np.around(x * xc / 2. + xc).astype(np.long)
                i = np.around(yc + y * yc).astype(np.long)
        return i, j

    xy2ij.__doc__ = SphericalProj.xy2ij.__doc__ % (name, name)

    def ij2xy(self, i=None, j=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for " "this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        ysize = xsize // 2
        xc, yc = (xsize - 1.) / 2., (ysize - 1.) / 2.
        if i is None and j is None:
            idx = np.outer(np.arange(ysize), np.ones(xsize))
            y = (idx - yc) / yc
            idx = np.outer(np.ones(ysize), np.arange(xsize))
            x = 2. * (idx - xc) / xc
            mask = x ** 2 / 4. + y ** 2 > 1.
            if not mask.any():
                mask = np.ma.nomask
            x = np.ma.array(x, mask=mask)
            y = np.ma.array(y, mask=mask)
        elif i is not None and j is not None:
            y = (np.asarray(i) - yc) / yc
            x = 2. * (np.asarray(j) - xc) / xc
            if x ** 2 / 4. + y ** 2 > 1.:
                x, y = np.nan, np.nan
        elif i is not None and j is None:
            i, j = i
            y = (np.asarray(i) - yc) / yc
            x = 2. * (np.asarray(j) - xc) / xc
            if x ** 2 / 4. + y ** 2 > 1.:
                x, y = np.nan, np.nan
        else:
            raise TypeError("i and j must be both given or both not given")
        return x, y

    ij2xy.__doc__ = SphericalProj.ij2xy.__doc__ % (name, name)

    def get_extent(self):
        return (-2.0, 2.0, -1.0, 1.0)

    @staticmethod
    def __initialise_data():
        if len(MollweideProj.__molldata) == 0:
            X = (np.arange(1., 180., 1.) - 90.) * dtor
            Y = MollweideProj.__findRoot(
                MollweideProj.__fmoll, MollweideProj.__dfmoll, X.copy(), X, niter=10
            )
            X = np.concatenate([[-pi / 2], X, [pi / 2]])
            Y = np.concatenate([[-pi / 2], Y, [pi / 2]])
            MollweideProj.__molldata.append(X)
            MollweideProj.__molldata.append(Y)
        return

    @staticmethod
    def __findRoot(f, df, x0, argsf=None, argsdf=None, niter=100):
        x = x0
        niter = min(abs(niter), 1000)
        i = 0
        while i < niter:
            dx = -f(x, argsf) / df(x, argsdf)
            x += dx
            i += 1
        return x

    @staticmethod
    def __fmoll(x, args):
        return 2. * x + np.sin(2. * x) - pi * np.sin(args)

    @staticmethod
    def __dfmoll(x, args):
        return 2. * (1. + np.cos(2. * x))

    @staticmethod
    def __lininterp(X, Y, x):
        idx = X.searchsorted(x)
        y = Y[idx - 1] + (Y[idx] - Y[idx - 1]) / (X[idx] - X[idx - 1]) * (
            x - X[idx - 1]
        )
        return y


class CartesianProj(SphericalProj):
    """This class provides class methods for Cartesian projection.
    """

    name = "Cartesian"

    def __init__(
        self,
        rot=None,
        coord=None,
        xsize=800,
        ysize=None,
        lonra=None,
        latra=None,
        **kwds
    ):
        super(CartesianProj, self).__init__(
            rot=rot,
            coord=coord,
            xsize=xsize,
            ysize=ysize,
            lonra=lonra,
            latra=latra,
            **kwds
        )

    def set_proj_plane_info(self, xsize, ysize, lonra, latra):
        if lonra is None:
            lonra = [-180., 180.]
        else:
            # shift lonra[1] into the range [lonra[0], lonra[0]+360]
            lonra_span = np.mod(lonra[1] - lonra[0], 360)
            if lonra_span == 0:
                lonra_span = 360
            lonra[1] = lonra[0] + lonra_span
        if latra is None:
            latra = [-90., 90.]
        if (
            len(lonra) != 2
            or len(latra) != 2
            or latra[0] < -90
            or latra[1] > 90
            or latra[0] >= latra[1]
        ):
            raise TypeError(
                "Wrong argument lonra or latra. Must be lonra=[a,b],latra=[c,d] "
                "c<d, c>=-90, d<=+90"
            )
        lonra = self._flip * np.float64(lonra)[:: self._flip]
        latra = np.float64(latra)
        xsize = np.long(xsize)
        if ysize is None:
            ratio = (latra[1] - latra[0]) / (lonra[1] - lonra[0])
            ysize = np.long(round(ratio * xsize))
        else:
            ysize = np.long(ysize)
            ratio = float(ysize) / float(xsize)
        super(CartesianProj, self).set_proj_plane_info(
            xsize=xsize, lonra=lonra, latra=latra, ysize=ysize, ratio=ratio
        )

    def vec2xy(self, vx, vy=None, vz=None, direct=False):
        if not direct:
            theta, phi = R.vec2dir(self.rotator(vx, vy, vz))
        else:
            theta, phi = R.vec2dir(vx, vy, vz)
        flip = self._flip
        # set phi in [-pi,pi]
        x = flip * ((phi + pi) % (2 * pi) - pi)
        x /= dtor  # convert in degree
        y = pi / 2. - theta
        y /= dtor  # convert in degree
        return x, y

    vec2xy.__doc__ = SphericalProj.vec2xy.__doc__ % (name, name)

    def xy2vec(self, x, y=None, direct=False):
        if y is None:
            x, y = np.asarray(x)
        else:
            x, y = np.asarray(x), np.asarray(y)
        flip = self._flip
        theta = pi / 2. - y * dtor  # convert in radian
        phi = flip * x * dtor  # convert in radian
        # dir2vec does not support 2d arrays, so first use flatten and then
        # reshape back to previous shape
        if not direct:
            vec = self.rotator.I(R.dir2vec(theta.flatten(), phi.flatten()))
        else:
            vec = R.dir2vec(theta.flatten(), phi.flatten())
        vec = [v.reshape(theta.shape) for v in vec]
        return vec

    xy2vec.__doc__ = SphericalProj.xy2vec.__doc__ % (name, name)

    def ang2xy(self, theta, phi=None, lonlat=False, direct=False):
        return self.vec2xy(R.dir2vec(theta, phi, lonlat=lonlat), direct=direct)

    ang2xy.__doc__ = SphericalProj.ang2xy.__doc__ % (name, name)

    def xy2ang(self, x, y=None, lonlat=False, direct=False):
        vec = self.xy2vec(x, y, direct=direct)
        return R.vec2dir(vec, lonlat=lonlat)

    xy2ang.__doc__ = SphericalProj.xy2ang.__doc__ % (name, name)

    def xy2ij(self, x, y=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for " "this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        ysize = int(self.arrayinfo["ysize"])
        lonra = self.arrayinfo["lonra"]
        latra = self.arrayinfo["latra"]
        if y is None:
            x, y = np.asarray(x)
        else:
            x, y = np.asarray(x), np.asarray(y)
        j = np.around((x - lonra[0]) / (lonra[1] - lonra[0]) * (xsize - 1)).astype(
            np.int64
        )
        i = np.around((y - latra[0]) / (latra[1] - latra[0]) * (ysize - 1)).astype(
            np.int64
        )
        if len(x.shape) > 0:
            mask = (i < 0) | (i >= ysize) | (j < 0) | (j >= xsize)
            if not mask.any():
                mask = np.ma.nomask
            j = np.ma.array(j, mask=mask)
            i = np.ma.array(i, mask=mask)
        else:
            if j < 0 or j >= xsize or i < 0 or i >= ysize:
                i = j = None
        return i, j

    xy2ij.__doc__ = SphericalProj.xy2ij.__doc__ % (name, name)

    def ij2xy(self, i=None, j=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for " "this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        ysize = int(self.arrayinfo["ysize"])
        lonra = self.arrayinfo["lonra"]
        latra = self.arrayinfo["latra"]
        if i is not None and j is None:
            i, j = np.asarray(i)
        elif i is not None and j is not None:
            i, j = np.asarray(i), np.asarray(j)
        if i is None and j is None:
            idx = np.outer(np.arange(ysize), np.ones(xsize))
            y = (float(latra[1] - latra[0]) / (ysize - 1.)) * idx
            y += latra[0]
            idx = np.outer(np.ones(ysize), np.arange(xsize))
            x = float(lonra[1] - lonra[0]) / (xsize - 1.) * idx
            x += lonra[0]
            x = np.ma.array(x)
            y = np.ma.array(y)
        elif i is not None and j is not None:
            y = (float(latra[1] - latra[0]) / (ysize - 1)) * i
            y += latra[0]
            x = (float(lonra[1] - lonra[0]) / (xsize - 1)) * j
            x += lonra[0]
            if len(i.shape) > 0:
                mask = (x < -180) | (x > 180) | (y < -90) | (y > 90)
                if not mask.any():
                    mask = np.ma.nomask
                x = np.ma.array(x, mask=mask)
                y = np.ma.array(y, mask=mask)
            else:
                if x < -180 or x > 180 or y < -90 or y > 90:
                    x = y = np.nan
        else:
            raise TypeError("i and j must be both given or both not given")
        return x, y

    ij2xy.__doc__ = SphericalProj.ij2xy.__doc__ % (name, name)

    def get_extent(self):
        lonra = self.arrayinfo["lonra"]
        latra = self.arrayinfo["latra"]
        return (lonra[0], lonra[1], latra[0], latra[1])

    get_extent.__doc__ = SphericalProj.get_extent.__doc__

    def get_fov(self):
        xsize = int(self.arrayinfo["xsize"])
        ysize = int(self.arrayinfo["ysize"])
        v1 = np.asarray(self.xy2vec(self.ij2xy(0, 0), direct=True))
        v2 = np.asarray(self.xy2vec(self.ij2xy(ysize - 1, xsize - 1), direct=True))
        a = np.arccos((v1 * v2).sum())
        return 2 * a

    #    def get_fov(self):
    #        lonra = self.arrayinfo['lonra']
    #        latra = self.arrayinfo['latra']
    #        return np.sqrt((lonra[1]-lonra[0])**2+(latra[1]-latra[0])**2)

    def get_center(self, lonlat=False):
        lonra = self.arrayinfo["lonra"]
        latra = self.arrayinfo["latra"]
        xc = 0.5 * (lonra[1] + lonra[0])
        yc = 0.5 * (latra[1] + latra[0])
        return self.xy2ang(xc, yc, lonlat=lonlat)

    get_center.__doc__ = SphericalProj.get_center.__doc__


class OrthographicProj(SphericalProj):
    """This class provides methods for orthographic projection
    """

    name = "Orthographic"

    def __init__(self, rot=None, coord=None, xsize=800, half_sky=False, **kwds):
        super(OrthographicProj, self).__init__(
            rot=rot, coord=coord, xsize=xsize, half_sky=half_sky, **kwds
        )

    def set_proj_plane_info(self, xsize, half_sky):
        super(OrthographicProj, self).set_proj_plane_info(
            xsize=xsize, half_sky=half_sky
        )

    def vec2xy(self, vx, vy=None, vz=None, direct=False):
        if not direct:
            theta, phi = R.vec2dir(self.rotator(vx, vy, vz))
        else:
            theta, phi = R.vec2dir(vx, vy, vz)
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for" " this projector"
            )
        half_sky = self.arrayinfo["half_sky"]
        flip = self._flip
        # set phi in [-pi,pi]
        phi = flip * (phi + pi) % (2 * pi) - pi
        lat = pi / 2. - theta
        x = np.cos(lat) * np.sin(phi)
        if not half_sky:
            x -= 1.0
        y = np.sin(lat)
        # unfold back of sphere
        cosc = np.cos(lat) * np.cos(phi)
        if np.any(cosc < 0):
            hmask = cosc < 0
            if hasattr(x, "__len__"):
                if half_sky:
                    x[hmask] = np.nan
                else:
                    x[hmask] *= -1
            elif hmask:
                if half_sky:
                    x = np.nan
                else:
                    x *= -1
        if half_sky:
            mask = np.asarray(x) ** 2 + np.asarray(y) ** 2 > 1.0
        else:
            mask = (np.mod(np.asarray(x) + 2.0, 2.0) - 1.0) ** 2 + np.asarray(
                y
            ) ** 2 > 1.0
        if mask.any():
            if not hasattr(x, "__len__"):
                x = np.nan
                y = np.nan
            else:
                x[mask] = np.nan
                y[mask] = np.nan
        return x, y

    vec2xy.__doc__ = SphericalProj.vec2xy.__doc__ % (name, name)

    def xy2vec(self, x, y=None, direct=False):
        if y is None:
            x, y = x
        if hasattr(x, "__len__"):
            x, y = np.asarray(x), np.asarray(y)
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for" " this projector"
            )
        half_sky = self.arrayinfo["half_sky"]
        flip = self._flip
        # re-fold back of sphere
        mask = None
        if not half_sky:
            if hasattr(x, "__len__"):
                if np.any(x > 0.0):
                    mask = x > 0.0
                    x[mask] *= -1
            elif x > 0:
                mask = 0
                x = -x
            x += 1.0
        r = np.sqrt(x ** 2 + y ** 2)
        if hasattr(r, "__len__"):
            r[r > 1] = np.nan
        elif r > 1:
            r = np.nan
        c = np.arcsin(r)
        if hasattr(y, "__len__"):
            y[np.abs(y) > 1] = np.nan
        elif np.abs(y) > 1:
            y = np.nan
        lat = np.arcsin(y)
        phi = np.arctan2(x, np.cos(c))
        phi *= flip
        if not mask is None:
            if hasattr(phi, "__len__"):
                phi[mask] = pi - phi[mask]
            else:
                phi = pi - phi
        theta = pi / 2. - lat
        vec = R.dir2vec(theta, phi)
        if not direct:
            return self.rotator.I(vec)
        else:
            return vec

    xy2vec.__doc__ = SphericalProj.xy2vec.__doc__ % (name, name)

    def ang2xy(self, theta, phi=None, lonlat=False, direct=False):
        return self.vec2xy(R.dir2vec(theta, phi, lonlat=lonlat), direct=direct)

    ang2xy.__doc__ = SphericalProj.ang2xy.__doc__ % (name, name)

    def xy2ang(self, x, y=None, lonlat=False, direct=False):
        return R.vec2dir(self.xy2vec(x, y, direct=direct), lonlat=lonlat)

    xy2ang.__doc__ = SphericalProj.xy2ang.__doc__ % (name, name)

    def xy2ij(self, x, y=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for" " this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        half_sky = self.arrayinfo["half_sky"]
        if half_sky:
            ratio = 1
        else:
            ratio = 2
        ysize = xsize // ratio
        if y is None:
            x, y = np.asarray(x)
        else:
            x, y = np.asarray(x), np.asarray(y)
        xc, yc = (xsize - 1.) / 2., (ysize - 1.) / 2.
        if hasattr(x, "__len__"):
            if half_sky:
                mask = x ** 2 + y ** 2 > 1.0
            else:
                mask = (np.mod(x + 2.0, 2.0) - 1.0) ** 2 + y ** 2 > 1.0
            if not mask.any():
                mask = np.ma.nomask
            j = np.ma.array(np.around(x * xc / ratio + xc).astype(np.long), mask=mask)
            i = np.ma.array(np.around(yc + y * yc).astype(np.long), mask=mask)
        else:
            if (half_sky and x ** 2 + y ** 2 > 1.0) or (
                not half_sky and (np.mod(x + 2.0, 2.0) - 1.0) ** 2 + y ** 2 > 1.0
            ):
                i, j, = np.nan, np.nan
            else:
                j = np.around(x * xc / ratio + xc).astype(np.long)
                i = np.around(yc + y * yc).astype(np.long)
        return i, j

    xy2ij.__doc__ = SphericalProj.xy2ij.__doc__ % (name, name)

    def ij2xy(self, i=None, j=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for" " this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        half_sky = self.arrayinfo["half_sky"]
        if half_sky:
            ratio = 1
        else:
            ratio = 2
        ysize = xsize // ratio
        xc, yc = (xsize - 1.) / 2., (ysize - 1.) / 2.
        if i is None and j is None:
            idx = np.outer(np.arange(ysize), np.ones(xsize))
            y = (idx - yc) / yc
            idx = np.outer(np.ones(ysize), np.arange(xsize))
            x = ratio * (idx - xc) / xc
        elif i is not None and j is not None:
            y = (np.asarray(i) - yc) / yc
            x = ratio * (np.asarray(j) - xc) / xc
            # if np.mod(x,1.0)**2+y**2 > 1.0: x,y=np.nan,np.nan
        elif i is not None and j is None:
            i, j = i
            y = (np.asarray(i) - yc) / yc
            x = ratio * (np.asarray(j) - xc) / xc
            # if np.mod(x,1.0)**2.+y**2 > 1.: x,y=np.nan,np.nan
        else:
            raise TypeError("i and j must be both given or both not given")
        if half_sky:
            mask = x ** 2 + y ** 2 > 1.
        else:
            mask = (np.mod(x + 2.0, 2.0) - 1.0) ** 2 + y ** 2 > 1.
        if not mask.any():
            mask = np.ma.nomask
        x = np.ma.array(x, mask=mask)
        y = np.ma.array(y, mask=mask)
        if len(x) == 0:
            x = x[0]
        if len(y) == 0:
            y = y[0]
        return x, y

    ij2xy.__doc__ = SphericalProj.ij2xy.__doc__ % (name, name)

    def get_extent(self):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for" " this projector"
            )
        half_sky = self.arrayinfo["half_sky"]
        if half_sky:
            ratio = 1.0
        else:
            ratio = 2.0
        return (-ratio, ratio, -1.0, 1.0)

    get_extent.__doc__ = SphericalProj.get_extent.__doc__


class AzimuthalProj(SphericalProj):
    """This class provides methods for Lambert azimuthal equal-area projection and
    azimuthal equidistant projection
    """

    name = "Azimuthal"

    def __init__(
        self,
        rot=None,
        coord=None,
        xsize=None,
        ysize=None,
        reso=None,
        lamb=None,
        half_sky=None,
        **kwds
    ):
        super(AzimuthalProj, self).__init__(
            rot=rot,
            coord=coord,
            xsize=xsize,
            ysize=ysize,
            reso=reso,
            lamb=lamb,
            half_sky=half_sky,
            **kwds
        )

    def set_proj_plane_info(
        self, xsize=800, ysize=None, reso=1.5, lamb=True, half_sky=False
    ):
        if xsize is None:
            xsize = 800
        if ysize is None:
            ysize = xsize
        if reso is None:
            reso = 1.5
        if lamb is None:
            lamb = True
        if half_sky is None:
            half_sky = False
        super(AzimuthalProj, self).set_proj_plane_info(
            xsize=xsize, ysize=ysize, reso=reso, lamb=lamb, half_sky=half_sky
        )

    def vec2xy(self, vx, vy=None, vz=None, direct=False):
        if not direct:
            theta, phi = R.vec2dir(self.rotator(vx, vy, vz))
        else:
            theta, phi = R.vec2dir(vx, vy, vz)
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for" " this projector"
            )
        flip = self._flip
        lamb = self.arrayinfo["lamb"]
        half_sky = self.arrayinfo["half_sky"]
        # set phi in [-pi,pi]
        phi = flip * ((phi + pi) % (2 * pi) - pi)
        lat = pi / 2. - theta
        if lamb:
            kprime = np.sqrt(2. / (1. + np.cos(lat) * np.cos(phi)))
        else:
            c = np.arccos(np.cos(lat) * np.cos(phi))
            kprime = c / np.sin(c)
        x = kprime * np.cos(lat) * np.sin(phi)
        y = kprime * np.sin(lat)
        if lamb:
            r2max = 4.
        else:
            r2max = pi ** 2
        if half_sky:
            if lamb:
                r2max /= 2.
            else:
                r2max /= 4.
        mask = np.asarray(x) ** 2 + np.asarray(y) ** 2 > r2max
        if not hasattr(x, "__len__"):
            if mask is not np.ma.nomask:
                return np.nan, np.nan
        else:
            w = np.where(mask)
            x[w] = np.nan
            y[w] = np.nan
        return x, y

    vec2xy.__doc__ = SphericalProj.vec2xy.__doc__ % (name, name)

    def xy2vec(self, x, y=None, direct=False):
        if y is None:
            x, y = x
        if hasattr(x, "__len__"):
            x, y = np.asarray(x), np.asarray(y)
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for" " this projector"
            )
        flip = self._flip
        lamb = self.arrayinfo["lamb"]
        half_sky = self.arrayinfo["half_sky"]
        if lamb:
            r2max = 4.
        else:
            r2max = pi ** 2
        if half_sky:
            if lamb:
                r2max /= 2.
            else:
                r2max /= 4.
        mask = np.asarray(x) ** 2 + np.asarray(y) ** 2 > r2max
        w = np.where(mask == False)
        if not mask.any():
            mask = np.ma.nomask
        if not hasattr(x, "__len__"):
            if mask is not np.ma.nomask:
                return np.nan, np.nan, np.nan
            else:
                rho = np.sqrt(x ** 2 + y ** 2)
                if lamb:
                    c = 2. * np.arcsin(rho / 2.)
                else:
                    c = rho
                lat = np.arcsin(y * np.sin(c) / rho)
                phi = np.arctan2(x * np.sin(c), (rho * np.cos(c)))
                phi *= flip
                vec = R.dir2vec(pi / 2. - lat, phi)
                if not direct:
                    return self.rotator.I(vec)
                else:
                    return vec
        else:
            vec = (
                np.zeros(x.shape) + np.nan,
                np.zeros(x.shape) + np.nan,
                np.zeros(x.shape) + np.nan,
            )
            rho = np.sqrt(x[w] ** 2 + y[w] ** 2)
            if lamb:
                c = 2. * np.arcsin(rho / 2.)
            else:
                c = rho
            lat = np.arcsin(y[w] * np.sin(c) / rho)
            phi = np.arctan2(x[w] * np.sin(c), (rho * np.cos(c)))
            phi *= flip
            vec[0][w] = np.cos(phi) * np.cos(lat)
            vec[1][w] = np.sin(phi) * np.cos(lat)
            vec[2][w] = np.sin(lat)
            if not direct:
                return self.rotator.I(vec)
            else:
                return vec

    xy2vec.__doc__ = SphericalProj.xy2vec.__doc__ % (name, name)

    def ang2xy(self, theta, phi=None, lonlat=False, direct=False):
        return self.vec2xy(R.dir2vec(theta, phi, lonlat=lonlat), direct=direct)

    ang2xy.__doc__ = SphericalProj.ang2xy.__doc__ % (name, name)

    def xy2ang(self, x, y=None, lonlat=False, direct=False):
        return R.vec2dir(self.xy2vec(x, y, direct=direct), lonlat=lonlat)

    xy2ang.__doc__ = SphericalProj.xy2ang.__doc__ % (name, name)

    def xy2ij(self, x, y=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for " "this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        ysize = int(self.arrayinfo["ysize"])
        reso = self.arrayinfo["reso"]
        lamb = self.arrayinfo["lamb"]
        half_sky = self.arrayinfo["half_sky"]
        if lamb:
            r2max = 4.
        else:
            r2max = pi ** 2
        if half_sky:
            if lamb:
                r2max /= 2.
            else:
                r2max /= 4.
        if y is None:
            x, y = x
        dx = reso / 60. * dtor
        xc, yc = 0.5 * (xsize - 1), 0.5 * (ysize - 1)
        if hasattr(x, "__len__"):
            mask = x ** 2 + y ** 2 > r2max
            if not mask.any():
                mask = np.ma.nomask
            j = np.ma.array(np.around(xc + x / dx).astype(np.long), mask=mask)
            i = np.ma.array(np.around(yc + y / dx).astype(np.long), mask=mask)
        else:
            if x ** 2 + y ** 2 > r2max:
                i, j, = np.nan, np.nan
            else:
                j = np.around(xc + x / dx).astype(np.long)
                i = np.around(yc + y / dx).astype(np.long)
        return i, j

    xy2ij.__doc__ = SphericalProj.xy2ij.__doc__ % (name, name)

    def ij2xy(self, i=None, j=None):
        if self.arrayinfo is None:
            raise TypeError(
                "No projection plane array information defined for " "this projector"
            )
        xsize = int(self.arrayinfo["xsize"])
        ysize = int(self.arrayinfo["ysize"])
        reso = self.arrayinfo["reso"]
        lamb = self.arrayinfo["lamb"]
        half_sky = self.arrayinfo["half_sky"]
        dx = reso / 60. * dtor
        xc, yc = 0.5 * (xsize - 1), 0.5 * (ysize - 1)
        if lamb:
            r2max = 4.
        else:
            r2max = pi ** 2
        if half_sky:
            if lamb:
                r2max /= 2.
            else:
                r2max /= 4.
        if i is None and j is None:
            idx = np.outer(np.arange(ysize), np.ones(xsize))
            y = (idx - yc) * dx
            idx = np.outer(np.ones(ysize), np.arange(xsize))
            x = (idx - xc) * dx
        elif i is not None and j is not None:
            y = (np.asarray(i) - yc) * dx
            x = (np.asarray(j) - xc) * dx
        elif i is not None and j is None:
            i, j = i
            y = (np.asarray(i) - yc) * dx
            x = (np.asarray(j) - xc) * dx
        else:
            raise TypeError("i and j must be both given or both not given")
        if hasattr(x, "__len__"):
            mask = x ** 2 + y ** 2 > r2max
            if not mask.any():
                mask = np.ma.nomask
            x = np.ma.array(x, mask=mask)
            y = np.ma.array(y, mask=mask)
        else:
            if x ** 2 + y ** 2 > r2max:
                x, y = np.nan, np.nan
        return x, y

    ij2xy.__doc__ = SphericalProj.ij2xy.__doc__ % (name, name)

    def get_extent(self):
        xsize = int(self.arrayinfo["xsize"])
        ysize = int(self.arrayinfo["ysize"])
        reso = self.arrayinfo["reso"]
        dx = reso / 60.0 * dtor
        xc, yc = 0.5 * (xsize - 1), 0.5 * (ysize - 1)
        left = -xc * dx
        bottom = -yc * dx
        right = (xsize - 1 - xc) * dx
        top = (ysize - 1 - yc) * dx
        return (left, right, bottom, top)

    get_extent.__doc__ = SphericalProj.get_extent.__doc__

    def get_fov(self):
        half_sky = self.arrayinfo["half_sky"]
        vx, vy, vz = self.xy2vec(self.ij2xy(0, 0), direct=True)
        a = np.arccos(vx)
        if np.isfinite(a):
            return 2. * a
        else:
            if half_sky:
                return pi
            else:
                return 2. * pi



import numpy as np

import unittest

import healpy as hp
from astropy.utils import data

import warnings

# disable new order warnings in tests
warnings.filterwarnings("ignore")


class TestMap2Alm(unittest.TestCase):

    def setUp(self):
        self.nside = 64
        self.lmax = 96
        alm_size = 4753
        np.random.seed(123)
        self.input_alm = np.ones(alm_size, dtype=np.complex)
        self.m = hp.alm2map(self.input_alm, nside=self.nside, lmax=self.lmax)

    def test_astropy_download_file(self):
        data.conf.dataurl = "https://healpy.github.io/healpy-data/"
        print(
            data.get_pkg_data_filename(
                "full_weights/healpix_full_weights_nside_0032.fits", package="healpy"
            )
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Bootstrap setuptools installation. We require setuptools >= 3.2 because of a
# bug in earlier versions regarding C++ sources generated with Cython. See:
#    https://pypi.python.org/pypi/setuptools/3.6#id171
try:
    import pkg_resources

    pkg_resources.require("setuptools >= 3.2")
except pkg_resources.ResolutionError:
    from ez_setup import use_setuptools

    use_setuptools()

import os
import errno
import fnmatch
import sys
import shlex
from distutils.sysconfig import get_config_var, get_config_vars
from subprocess import check_output, CalledProcessError, check_call
from setuptools import setup
from setuptools.dist import Distribution
from distutils.command.build_clib import build_clib
from distutils.errors import DistutilsExecError
from distutils.dir_util import mkpath
from distutils.file_util import copy_file
from distutils import log

# Apple switched default C++ standard libraries (from gcc's libstdc++ to
# clang's libc++), but some pre-packaged Python environments such as Anaconda
# are built against the old C++ standard library. Luckily, we don't have to
# actually detect which C++ standard library was used to build the Python
# interpreter. We just have to propagate MACOSX_DEPLOYMENT_TARGET from the
# configuration variables to the environment.
#
# This workaround fixes <https://github.com/healpy/healpy/issues/151>.
if (
    get_config_var("MACOSX_DEPLOYMENT_TARGET")
    and not "MACOSX_DEPLOYMENT_TARGET" in os.environ
):
    os.environ["MACOSX_DEPLOYMENT_TARGET"] = get_config_var("MACOSX_DEPLOYMENT_TARGET")


# If the Cython-generated C++ files are absent, then fetch and install Cython
# as an egg. If the Cython-generated files are present, then only use Cython if
# a sufficiently new version of Cython is already present on the system.
cython_require = "Cython >= 0.16"
log.info("checking if Cython-generated files have been built")
try:
    open("healpy/src/_query_disc.cpp")
except IOError:
    log.info("Cython-generated files are absent; installing Cython locally")
    Distribution().fetch_build_eggs(cython_require)
else:
    log.info("Cython-generated files are present")
try:
    log.info("Checking for %s", cython_require)
    pkg_resources.require(cython_require)
except pkg_resources.ResolutionError:
    log.info("%s is not installed; not using Cython")
    from setuptools.command.build_ext import build_ext
    from setuptools import Extension
else:
    log.info("%s is installed; using Cython")
    from Cython.Distutils import build_ext, Extension


class build_external_clib(build_clib):
    """Subclass of Distutils' standard build_clib subcommand. Adds support for
    libraries that are installed externally and detected with pkg-config, with
    an optional fallback to build from a local configure-make-install style
    distribution."""

    def __init__(self, dist):
        build_clib.__init__(self, dist)
        self.build_args = {}

    def autotools_path(self):
        """
        Install Autotools locally if we are building on Read The Docs, because
        we will be building from git and need to generate the healpix_cxx build
        system.
        """
        on_rtd = os.environ.get("READTHEDOCS", None) == "True"
        if not on_rtd:
            return None

        log.info("checking if autotools is installed")
        try:
            check_output(["autoreconf", "--version"])
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            log.info("autotools is not installed")
        else:
            log.info("autotools is already installed")
            return None

        urls = [
            "https://ftp.gnu.org/gnu/m4/m4-1.4.17.tar.gz",
            "https://ftp.gnu.org/gnu/libtool/libtool-2.4.6.tar.gz",
            "https://ftp.gnu.org/gnu/autoconf/autoconf-2.69.tar.gz",
            "https://ftp.gnu.org/gnu/automake/automake-1.15.tar.gz",
            "https://pkg-config.freedesktop.org/releases/pkg-config-0.29.1.tar.gz",
        ]

        # Use a subdirectory of build_temp as the build directory.
        build_temp = os.path.realpath(self.build_temp)
        prefix = os.path.join(build_temp, "autotools")
        mkpath(prefix)

        env = dict(os.environ)
        path = os.path.join(prefix, "bin")
        try:
            path += ":" + env["PATH"]
        except KeyError:
            pass
        env["PATH"] = path

        # Check again if autotools is available, now that we have added the
        # temporary path.
        try:
            check_output(["autoreconf", "--version"], env=env)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            log.info("building autotools from source")
        else:
            log.info("using autotools built from source")
            return path

        # Otherwise, build from source.
        for url in urls:
            _, _, tarball = url.rpartition("/")
            pkg_version = tarball.replace(".tar.gz", "")
            log.info("downloading %s", url)
            check_call(["curl", "-O", url], cwd=build_temp)
            log.info("extracting %s", tarball)
            check_call(["tar", "-xzf", tarball], cwd=build_temp)
            cwd = os.path.join(build_temp, pkg_version)
            log.info("configuring %s", pkg_version)
            check_call(
                ["./configure", "--prefix", prefix, "--with-internal-glib"],
                env=env,
                cwd=cwd,
            )
            log.info("making %s", pkg_version)
            check_call(["make", "install"], env=env, cwd=cwd)
        return path

    def env(self):
        """Construct an environment dictionary suitable for having pkg-config
        pick up .pc files in the build_clib directory."""
        # Test if pkg-config is present. If not, fall back to pykg-config.
        try:
            env = self._env
        except AttributeError:
            env = dict(os.environ)

            path = self.autotools_path()
            if path is not None:
                env["PATH"] = path

            try:
                check_output(["pkg-config", "--version"])
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
                log.warn("pkg-config is not installed, falling back to pykg-config")
                env["PKG_CONFIG"] = (
                    sys.executable + " " + os.path.abspath("run_pykg_config.py")
                )
            else:
                env["PKG_CONFIG"] = "pkg-config"

            build_clib = os.path.realpath(self.build_clib)
            pkg_config_path = (
                os.path.join(build_clib, "lib64", "pkgconfig")
                + ":"
                + os.path.join(build_clib, "lib", "pkgconfig")
            )
            try:
                pkg_config_path += ":" + env["PKG_CONFIG_PATH"]
            except KeyError:
                pass
            env["PKG_CONFIG_PATH"] = pkg_config_path

            self._env = env
        return env

    def pkgconfig(self, *packages):
        env = self.env()
        PKG_CONFIG = tuple(shlex.split(env["PKG_CONFIG"], posix=(os.sep == "/")))
        kw = {}
        index_key_flag = (
            (2, "--cflags-only-I", ("include_dirs",)),
            (0, "--cflags-only-other", ("extra_compile_args", "extra_link_args")),
            (2, "--libs-only-L", ("library_dirs", "runtime_library_dirs")),
            (2, "--libs-only-l", ("libraries",)),
            (0, "--libs-only-other", ("extra_link_args",)),
        )
        for index, flag, keys in index_key_flag:
            cmd = PKG_CONFIG + (flag,) + tuple(packages)
            log.debug("%s", " ".join(cmd))
            args = [
                token[index:].decode() for token in check_output(cmd, env=env).split()
            ]
            if args:
                for key in keys:
                    kw.setdefault(key, []).extend(args)
        return kw

    def finalize_options(self):
        """Run 'autoreconf -i' for any bundled libraries to generate the
        configure script."""
        build_clib.finalize_options(self)
        env = self.env()

        for lib_name, build_info in self.libraries:
            if "sources" not in build_info:
                log.info(
                    "checking if configure script for library '%s' exists", lib_name
                )
                if not os.path.exists(
                    os.path.join(build_info["local_source"], "configure")
                ):
                    log.info("running 'autoreconf -i' for library '%s'", lib_name)
                    check_call(
                        ["autoreconf", "-i"], cwd=build_info["local_source"], env=env
                    )

    def build_library(
        self,
        library,
        pkg_config_name,
        local_source=None,
        supports_non_srcdir_builds=True,
    ):
        log.info("checking if library '%s' is installed", library)
        try:
            build_args = self.pkgconfig(pkg_config_name)
            log.info("found '%s' installed, using it", library)
        except CalledProcessError:

            # If local_source is not specified, then immediately fail.
            if local_source is None:
                raise DistutilsExecError("library '%s' is not installed", library)

            log.info("building library '%s' from source", library)

            env = self.env()

            # Determine which compilers we are to use, and what flags.
            # This is based on what distutils.sysconfig.customize_compiler()
            # does, but that function has a problem that it doesn't produce
            # necessary (e.g. architecture) flags for C++ compilers.
            cc, cxx, opt, cflags = get_config_vars("CC", "CXX", "OPT", "CFLAGS")
            cxxflags = cflags

            if "CC" in env:
                cc = env["CC"]
            if "CXX" in env:
                cxx = env["CXX"]
            if "CFLAGS" in env:
                cflags = opt + " " + env["CFLAGS"]
            if "CXXFLAGS" in env:
                cxxflags = opt + " " + env["CXXFLAGS"]

            # Use a subdirectory of build_temp as the build directory.
            build_temp = os.path.realpath(os.path.join(self.build_temp, library))

            # Destination for headers and libraries is build_clib.
            build_clib = os.path.realpath(self.build_clib)

            # Create build directories if they do not yet exist.
            mkpath(build_temp)
            mkpath(build_clib)

            if not supports_non_srcdir_builds:
                self._stage_files_recursive(local_source, build_temp)

            # Run configure.
            cmd = [
                "/bin/sh",
                os.path.join(os.path.realpath(local_source), "configure"),
                "--prefix=" + build_clib,
                "--disable-shared",
                "--with-pic",
                "--disable-maintainer-mode",
            ]

            log.info("%s", " ".join(cmd))
            check_call(
                cmd,
                cwd=build_temp,
                env=dict(env, CC=cc, CXX=cxx, CFLAGS=cflags, CXXFLAGS=cxxflags),
            )

            # Run make install.
            cmd = ["make", "install"]
            log.info("%s", " ".join(cmd))
            check_call(cmd, cwd=build_temp, env=env)

            build_args = self.pkgconfig(pkg_config_name)

        return build_args
        # Done!

    @staticmethod
    def _list_files_recursive(path, skip=(".*", "*.o", "autom4te.cache")):
        """Yield paths to all of the files contained within the given path,
        following symlinks. If skip is a tuple of fnmatch()-style wildcard
        strings, skip any directory or filename matching any of the patterns in
        skip."""
        for dirpath, dirnames, filenames in os.walk(path, followlinks=True):
            if not any(
                any(fnmatch.fnmatch(p, s) for s in skip) for p in dirpath.split(os.sep)
            ):
                for filename in filenames:
                    if not any(fnmatch.fnmatch(filename, s) for s in skip):
                        yield os.path.join(dirpath, filename)

    @staticmethod
    def _stage_files_recursive(src, dest, skip=None):
        """Hard link or copy all of the files in the path src into the path dest.
        Subdirectories are created as needed, and files in dest are overwritten."""
        # Use hard links if they are supported on this system.
        if hasattr(os, "link"):
            link = "hard"
        elif hasattr(os, "symlink"):
            link = "sym"
        else:
            link = None

        for dirpath, dirnames, filenames in os.walk(src, followlinks=True):
            if not any(p.startswith(".") for p in dirpath.split(os.sep)):
                dest_dirpath = os.path.join(
                    dest, dirpath.split(src, 1)[1].lstrip(os.sep)
                )
                mkpath(dest_dirpath)
                for filename in filenames:
                    if not filename.startswith("."):
                        src_path = os.path.join(dirpath, filename)
                        dest_path = os.path.join(dest_dirpath, filename)
                        if not os.path.exists(dest_path):
                            copy_file(
                                os.path.join(dirpath, filename),
                                os.path.join(dest_dirpath, filename),
                            )

    def get_source_files(self):
        """Copied from Distutils' own build_clib, but modified so that it is not
        an error for a build_info dictionary to lack a 'sources' key. If there
        is no 'sources' key, then all files contained within the path given by
        the 'local_sources' value are returned."""
        self.check_library_list(self.libraries)
        filenames = []
        for (lib_name, build_info) in self.libraries:
            sources = build_info.get("sources")
            if sources is None or not isinstance(sources, (list, tuple)):
                sources = list(self._list_files_recursive(build_info["local_source"]))

            filenames.extend(sources)
        return filenames

    def build_libraries(self, libraries):
        # Build libraries that have no 'sources' key, accumulating the output
        # from pkg-config.
        for lib_name, build_info in libraries:
            if "sources" not in build_info:
                for key, value in self.build_library(lib_name, **build_info).items():
                    if key in self.build_args:
                        self.build_args[key].extend(value)
                    else:
                        self.build_args[key] = value

        # Use parent method to build libraries that have a 'sources' key.
        build_clib.build_libraries(
            self,
            (
                (lib_name, build_info)
                for lib_name, build_info in libraries
                if "sources" in build_info
            ),
        )


class custom_build_ext(build_ext):

    def finalize_options(self):
        build_ext.finalize_options(self)

        # Make sure that Numpy is importable
        # (does the same thing as setup_requires=['numpy'])
        self.distribution.fetch_build_eggs("numpy")
        # Prevent numpy from thinking it is still in its setup process:
        # See http://stackoverflow.com/questions/19919905
        from six.moves import builtins

        builtins.__NUMPY_SETUP__ = False

        # Add Numpy header search path path
        import numpy

        self.include_dirs.append(numpy.get_include())

    def run(self):
        # If we were asked to build any C/C++ libraries, add the directory
        # where we built them to the include path. (It's already on the library
        # path.)
        if self.distribution.has_c_libraries():
            self.run_command("build_clib")
            build_clib = self.get_finalized_command("build_clib")
            for key, value in build_clib.build_args.items():
                for ext in self.extensions:
                    if not hasattr(ext, key) or getattr(ext, key) is None:
                        setattr(ext, key, value)
                    else:
                        getattr(ext, key).extend(value)
        build_ext.run(self)


exec(open("healpy/version.py").read())

def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name="healpy",
    version=__version__,
    description="Healpix tools package for Python",
    long_description=readme(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: POSIX",
        "Programming Language :: C++",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    author="C. Rosset, A. Zonca",
    author_email="cyrille.rosset@apc.univ-paris-diderot.fr",
    url="http://github.com/healpy",
    packages=["healpy", "healpy.test"],
    libraries=[
        (
            "cfitsio",
            {
                "pkg_config_name": "cfitsio",
                "local_source": "cfitsio",
                "supports_non_srcdir_builds": False,
            },
        ),
        (
            "healpix_cxx",
            {
                "pkg_config_name": "healpix_cxx >= 3.40.0",
                "local_source": "healpixsubmodule/src/cxx/autotools",
            },
        ),
    ],
    py_modules=[
        "healpy.pixelfunc",
        "healpy.sphtfunc",
        "healpy.visufunc",
        "healpy.fitsfunc",
        "healpy.projector",
        "healpy.rotator",
        "healpy.projaxes",
        "healpy.version",
    ],
    cmdclass={"build_ext": custom_build_ext, "build_clib": build_external_clib},
    ext_modules=[
        Extension(
            "healpy._healpy_pixel_lib",
            sources=["healpy/src/_healpy_pixel_lib.cc"],
            language="c++",
        ),
        Extension(
            "healpy._healpy_sph_transform_lib",
            sources=["healpy/src/_healpy_sph_transform_lib.cc"],
            language="c++",
        ),
        Extension(
            "healpy._query_disc",
            ["healpy/src/_query_disc.pyx"],
            language="c++",
            cython_directives=dict(embedsignature=True),
        ),
        Extension(
            "healpy._sphtools",
            ["healpy/src/_sphtools.pyx"],
            language="c++",
            cython_directives=dict(embedsignature=True),
        ),
        Extension(
            "healpy._pixelfunc",
            ["healpy/src/_pixelfunc.pyx"],
            language="c++",
            cython_directives=dict(embedsignature=True),
        ),
    ],
    package_data={
        "healpy": [
            "data/*.fits",
            "data/totcls.dat",
            "test/data/*.fits",
            "test/data/*.fits.gz",
            "test/data/*.sh",
        ]
    },
    install_requires=["matplotlib", "numpy>=1.13", "six", "astropy", "scipy"],
    setup_requires=["pytest-runner", "six"],
    tests_require=["pytest", "pytest-cython"],
    test_suite="healpy",
    license="GPLv2",
)



import numpy as np
import pytest


def pytest_configure():
    """Set the Numpy print style to a fixed version to make doctest outputs
    reproducible."""
    try:
        np.set_printoptions(legacy="1.13")
    except TypeError:
        # On older versions of Numpy, the unrecognized 'legacy' option will
        # raise a TypeError.
        pass
